package Warsync::Server;

use strict;
use warnings;

use Warsync::Host;
use Warsync::Util;
use Warsync::Config;

use IPC::Open2;
use IO::Handle;
use File::Temp;
use DirHandle;

our @ISA = qw( Warsync::Host );

## here we have a regex that matches lines produced by rsync
## that don't signify that a change has been made, and at
## what verbosities to check for them.
## TODO: these regexes should be optimized by which statements
## occur most often.
my %rsync_regex = ();

## rsync -v
$rsync_regex{v} = qr/
    ^building \s+ file \s+ list
  | ^wrote \s+ \d+ \s+ bytes
  | ^total \s+ size \s+ is \s+ \d+
/x;

## rsync -vv
$rsync_regex{vv} = qr/
    ^including \s+ (?:file|directory)
  | \s+ is \s+ uptodate$
  | ^excluding \s+ (?:file|directory)
  | ^opening \s+ connection \s+ using
  | ^expand \s+ file_list \s+ to \s+ \d+ \s+ bytes
  | ^done$
  | ^deleting \s+ in \s+ \.
  | ^total: \s+ matches=\d+
  | $rsync_regex{v}
/x;

## rsync -vvv
$rsync_regex{vvv} = qr/
    ^add_exclude\(
  | ^make_file\(
  | ^(?:recv|send)_file_list \s+ done$
  | ^send_files \s+ starting$
  | ^server_recv\(
  | ^recv_file_name\(
  | ^received \s+ \d+ \s+ names$
  | ^get_local_name \s+ count=\d+
  | ^generator \s+ starting \s+ pid=
  | ^recv_generator\(
  | ^recv_files\(
  | ^(?:send|recv|generate)_files \s+ phase=
  | ^send \s+ files \s+ finished$
  | ^_exit_cleanup\(
  | $rsync_regex{vv}
/x;

## rsync -vvvv
$rsync_regex{vvvv} = qr/
    ^cmd=ssh
  | ^\[\d+\] \s+ i=\d+
  | ^file \s+ list \s+ sent$
  | ^server_recv\(
  | ^client_run \s+ waiting \s+ on \s+ \d+$
  | $rsync_regex{vvv}
/x;

sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;

    unless (-r '/etc/warsync/server-key' and -w '/etc/warsync/server.conf') {
        return undef;
    }

    mkdir '/etc/warsync/filepacks.d' unless -d '/etc/warsync/filepacks.d';

    my $config = Warsync::Config->new('/etc/warsync/server.conf')
        or fault("could not read server config: $!");

    return bless { config => $config, _display_queue => [] } => $class;
}

sub apt_conf
{
    my $self = shift->new;

    return '' unless -e '/etc/apt/apt.conf';

    open(my $fh, '/etc/apt/apt.conf')
        or fault("could not read /etc/apt/apt.conf");
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);

    return $slurp;
}

sub apt_preferences
{
    my $self = shift->new;

    return '' unless -e '/etc/apt/preferences';

    open(my $fh, '/etc/apt/preferences')
        or fault("could not read /etc/apt/preferences");
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);

    return $slurp;
}

sub apt_sources_list
{
    my $self = shift->new;

    open(my $fh, '/etc/apt/sources.list')
        or fault("could not read /etc/apt/sources.list");
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);

    return $slurp;
}

sub client
{
    my $self   = shift->new;
    my $client = shift;

    return undef unless $self->config->has_section("client-$client");

    return Warsync::Server::Client->new($self, $client);
}

sub config
{
    my $self = shift->new;

    return $self->{config};
}

sub display
{
    my $self = shift->new;

    ## if we are on a tty, something has changed, or want verbosity,
    ## display the output as appropriate.
    if (tty or $self->modified or $self->verbose > 1) {
        $self->SUPER::display(@_);
    }
    ## otherwise queue the message for a later time.
    else {
        push @{ $self->{'_display_queue'} }, [ @_ ];
    }
}

sub display_clear
{
    my $self = shift->new;
    
    $self->{'_display_queue'} = [];
    $self->{'_display_queue_block'} = undef;
}

sub filepacks
{
    my $self = shift->new;

    my $d = DirHandle->new('/etc/warsync/filepacks.d')
        or return undef;

    my (@packs) = sort grep { /^[^.]/ && -f "/etc/warsync/filepacks.d/$_" } $d->read;
                                                                                         $d->close;

    return @packs;
}

sub list_clients
{
    my $self = shift->new;

    return map  { $_ =~ s/^client-//; $_ }
           grep { $_ =~ m/^client-/  } $self->{config}->list_sections;
}

sub modified
{
    my $self = shift->new;
    if (@_) {
        ## set modified bit.
        $self->{_modified} = shift;
        ## if modified is true, dump currently queued output.
        my $dq = $self->{'_display_queue'};
        if ($self->{_modified} and @$dq) {
            $self->{'_display_queue_block'} = undef;
            while (my $out = shift @$dq) {
                $self->display(@$out);
            }
        }
        ## if modified now false, clear old queue if something there.
        else {
            if (defined $self->{'_display_queue_block'}) {
                splice @$dq, $self->{'_display_queue_block'};
            }
            else {
                $self->{'_display_queue_block'} = @$dq;
            }
        }
                    
        return $self;
    }
    return $self->{_modified} || 0;
}

sub replication
{
    my $self = shift->new;

    my $clients = $self->option('client') || [ $self->list_clients ];

    my @filepacks = @ARGV;

    if ($self->option('a')) {
        @filepacks = $self->filepacks;
    }

    ## they need to specify something to do.
    unless (@filepacks or $self->option('debian')) {
        Pod::Usage::pod2usage(1);
    }

    ## iterate through each client and replicate.
    foreach my $name (@$clients) {

        $self->display_clear;
        $self->display(1,
            "######################################################\n");

        my $client = $self->client($name);

        ## if we didn't get back a fqdn, this is a bad client name.
        unless ($client) {
            $self->display(0, $name, " is not a valid client\n");
            next;
        }

        my $dryrun = $self->option('n') ? ' (dry-run) ' : '';

        $self->display(1, "Replicating $name$dryrun...\n");

        ## check protocol of client to make sure it matches.
        if (my $prot_match = $client->protocol_match) {
            ## replicate filepacks.
            foreach my $filepack (@filepacks) {
                $client->replicate_filepack($filepack);
            }
            ## debian syncronization.
            if ($self->option('debian')) {
                $client->sync_debian_packages;
            }
        }
        elsif (defined $prot_match) {
            $self->display(0, "!! Client $name communication protocol differs.\n");
            $self->display(0, "!! Please manually upgrade client to version $Warsync::VERSION.\n");
        }
        else {
            $self->display(0, "!! Could not connect to client $name.\n");
        }
    }
}

## warsync-echo-name command. this simply echos the server's hostname.
sub rpc_echo_name
{
    my $self = shift->new;
    print 'name: ', $self->name, "\n";
}


#############################################################################
#############################################################################

package Warsync::Server::Client;

use Warsync::Util;

## Warsync::Server::Client->new($server, $client_name);
sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;

    my $server = shift;
    my $name   = shift;

    $self = bless {
        server => $server,
        name   => $name,
        config => $server->config->get_section("client-$name")
    } => $class;

    return $self;
}

sub base_command
{
    my $self = shift->new;
    my $cmd  = shift;

    my $v = $self->server->option('v');

    my $rcmd = 'warsync-';
    $rcmd   .= 'compress-'  if $self->compress;
    $rcmd   .= 'tty-'       if tty;
    $rcmd   .= 'dry-run-'   if $self->server->option('n');
    $rcmd   .= 'quiet-'     if $self->server->option('q');
    $rcmd   .= "verbose$v-" if $v;

    $rcmd   .= $cmd;

    return $rcmd;
}

sub system_call
{
    my $self = shift->new;

    $self->server->display(4, "s-syscall: @_\n");

    return system(@_) ? 0 : 1;
}

sub command
{
    my $self   = shift->new;
    my $cmd    = $self->base_command(shift);
    my @args   = @_;

    my @scmd = ( 'ssh', '-p', $self->port,
                        ($self->compress ? '-C' : () ),
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-T', $self->fqdn, $cmd, @args );

    open(my $ssh_h, '-|') || exec( @scmd )
        or die "error '$!' trying to exec: @scmd";

    return $ssh_h;
}

sub gui_command
{
    my $self   = shift->new;

    ## run as non-gui command if this is not a tty.
    unless (-t) {
        my $comh = $self->command(@_);
        print <$comh>;
        return close($comh)
    }

    my $cmd    = $self->base_command(shift);
    my @args   = @_;

    my @scmd = ( 'ssh', '-p', $self->port,
                        ($self->compress ? '-C' : () ),
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-qt', $self->fqdn, $cmd, @args );

    return $self->system_call(@scmd)
}

sub rw_command
{
    my $self   = shift->new;
    my $cmd    = $self->base_command(shift);
    my @args   = @_;

    my @scmd = ( 'ssh', '-p', $self->port,
                        ($self->compress ? '-C' : () ),
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-T', $self->fqdn, $cmd, @args );

    my ($rh, $wh);
    my $pid = IPC::Open2::open2($rh, $wh, @scmd);

    return $pid, $rh, $wh;
}

sub compress
{
    my $self = shift->new;

    if (@_) {
        $self->set_key('compress', shift);
        return $self;
    }
    ## the next two lines are needed to get work
    ## with perl 5.8's more restrictive taintmode.
    my $unsafe = $self->get_key('compress') || 0;
    my ($compress) = $unsafe =~ /(.)/;

    return $compress;
}

sub config
{
    my $self = shift->new;

    return $self->{config};
}

sub ignore_debs
{
    my $self   = shift->new;

    my @debs = split ' ', $self->get_key('ignore_debs') || '';
    return \@debs;
}

sub fqdn
{
    my $self = shift->new;

    if (@_) {
        $self->set_key('fqdn', is_fqdn(shift));
        return $self;
    }

    return is_fqdn($self->get_key('fqdn'));
}

sub get_key
{
    my $self = shift->new;

    return $self->config->get_key(shift);
}

sub set_key
{
    my $self = shift->new;

    return $self->config->set_key(shift, shift);
}

sub name
{
    my $self = shift->new;

    return $self->{name};
}

sub packages
{
    my $self = shift->new;

    my $cmd_h = $self->command('echo-package-list');
    my $cpkgs = parse_package_list($cmd_h);
    close($cmd_h);

    return $cpkgs;
}

sub port
{
    my $self = shift->new;

    if (@_) {
        my ($port) = shift =~ /(\d+)/;
        $self->set_key('port', $port);
        return $self;
    }

    my $unsafe = $self->get_key('port') || 22;
    my ($port) = $unsafe =~ /(\d+)/;

    return $port || 22;
}

sub protocol_match
{
    my $self = shift->new;

    my $com_h = $self->command('echo-protocol-version');

    chomp(my $ver = <$com_h> || 0);

    close($com_h)
        or return undef;

    return 1 if $ver == $Warsync::PROTOCOL;
    return 0;
}

sub replicate_filepack
{
    my $self = shift->new;
    my $unsafe_pack = shift;

    my $server = $self->server;

    $server->modified(0);

    my ($pack) = $unsafe_pack =~ /([-\w.]+)/;

    my $v = $server->option('v') || 0;

    $server->display(1, "* filepack $pack syncronization...\n");

    open(my $fh, "/etc/warsync/filepacks.d/$pack")
        or fault("Filepack $pack not found: $!");

    my @pre_cmd   = ();  # pre commands are run before replicatoin
    my @mod_cmd   = ();  # mod commands are run if files modified
    my @post_cmd  = ();  # post commands are run after replication

    my @rsync_opt = ();  # extra options to pass to rsync.

    my $base_dir  = '/'; # base directory to replicate

    my @include   = ();  # contains include order
    my %include   = ();  # used for include dup check

    my $line = 0;
    while (<$fh>) {
        ++$line;            # line number
        chomp;              # no newline
        s/#.*//;            # no comments
        s/^\s+//;           # no leading white
        s/\s+$//;           # no trailing white
        next unless length; # anything left?

        ## check for config setting.
        if ($line =~ /=/) {
            my ($var, $value) = split(/\s*=\s*/, $_, 2);
            $var = uc($var);
            if ($var eq 'PRECMD') {
                push @pre_cmd, $value;
            }
            elsif ($var eq 'POSTCMD') {
                push @post_cmd, $value;
            }
            elsif ($var eq 'MODCMD') {
                push @mod_cmd, $value;
            }
            #elsif ($var eq 'BASEDIR') {
            #    fault "BASEDIR specified on line $line of $pack ",
            #          "must be absolute" unless $value =~ m!^/!;
            #
            #    $value .= '/' unless $value =~ m/\/$/;
            #    $base_dir = $value;
            #}
            elsif ($var eq 'RSYNCOPT') {
                push @rsync_opt, $value;
            }
            else {
                fault "Unknown $var variable on line $line ",
                      "of $pack filepack.\n";
            }
            next;
        }

        ## read include line.
        my $inc = $_;

        ## check for list reset.
        if ($inc eq '!') {
            @include = ();
            %include = ();
            $server->display(2, "  - found '!' resetting filepack includes\n");
            next;
        }

        ## exclude if starts with -
        if ($inc =~ /^- /) {
            unless ($include{$inc}++) {
                push @include, $inc;
                (my $ex = $inc) =~ s/^- //;
                $server->display(2, "  - excluding $ex\n");
            }
            next;
        }

        ## remove forced include if it exists.
        $inc =~ s!^\+ !!;

        ## check if path is absolute
        my $abs = $inc =~ s!^/!!;

        ## split path into separate components.
        my @path = split('/', $inc);
        ## find the first without a wildcard.
        my @pre_path = ();
        while (my $c = shift @path) {
            ## path does not contain a wildcard,
            ## pre-include it.
            if ($c !~ m![\[\?\*\]]!) {
                push @pre_path, $c;
                my $path = join('/', @pre_path);
                ## add directory slash if not last component.
                $path .= '/' if @path;
                ## make absolute if it should be.
                $path  = "/$path" if $abs;
                ## add to includes if not already there.
                unless ($include{$path}++) {
                    push @include, $path;
                    $server->display(@path ? 3 : 2, "  - including $path\n");
                }
            }
            ## other wise if it contained a wildcard, put it back
            ## and dump out.
            else {
                unshift @path, $c;
                last;
            }
        }

        ## build final include
        my $path = join('/', @pre_path, @path);
        ## make absolute if necessary.
        $path = "/$path" if $abs;

        ## and finally do it.
        unless ($include{$path}++) {
            push @include, $path;
            $server->display(2, "  - including $path\n");
        }
    }

    close($fh);

    ## nothing to do if include is empty.
    unless (@include) {
        $server->display(1, "  - nothing included by filepack\n");
        return;
    }

    ## exclude everything not explicitly included.
    push @include, '- *' unless $include{'- *'}++;

    ## open tempfile to write include list.
    my ($tmpfh, $unsafe_tmpfn) = File::Temp::tempfile(UNLINK => 1);

    my ($tmpfn) = $unsafe_tmpfn =~ /([\/0-9A-Za-z]+)/;

    die "strange tmpfn $unsafe_tmpfn != $tmpfn"
        unless $tmpfn eq $unsafe_tmpfn;

    foreach my $inc (@include) {
        print $tmpfh $inc, "\n";
    }

    ## flush include list just to be safe.
    IO::Handle::flush($tmpfh);

    ## build ssh command to use for rsync communication.
    my @ssh   = ( 'ssh', '-p', $self->port,
                         '-o', 'NumberOfPasswordPrompts=0',
                         '-i', '/etc/warsync/server-key', );

    ## rsync needs to be a little verbose so we can get the output.
    ## but increase it's verbosity if warsync was called with -vv or more.
    my $vv = $v > 2 ? 'v' x ($v - 1) : 'v';
    ## but no need to have more than four.
    $vv = 'vvvv' if length($vv) > 4;

    my $dryrun = $server->option('n') ? ' (dry-run) ' : '';
    ($dryrun) = $dryrun =~ /(.*)/; ## untaint for perl 5.8.

    my $exclude_tmp = '--exclude=' . $tmpfn;
    my $include_tmp = '--include-from=' . $tmpfn;

    ## build rsync command.
    my @rsync = ( 'rsync', '--delete', '--rsh='.join(' ', @ssh),
                           $self->compress ? '-z'     : (),
                           $dryrun         ? "-an$vv" : "-a$vv",
                           @rsync_opt,
                           $exclude_tmp,
                           $include_tmp,
                           $base_dir, $self->fqdn.":$base_dir"
                );

    $server->display(1, "  - file replication$dryrun...\n");
    $server->display(3, "    # rsync: @rsync\n");

    open(my $rsync_h, '-|') || exec( @rsync )
        or die "error '$!' trying to exec: @rsync";

    while (<$rsync_h>) {
        chomp;
        s/^\s+//;           # no leading white
        s/\s+$//;           # no trailing white
        next unless length; # anything left?

        ## grep for lines that do not signify a change.
        if ($_ =~ /$rsync_regex{$vv}/) {
            $server->display(2, "    # ", $_, "\n");
        }
        ## change was made
        else {
            $server->modified(1);
            $server->display(1, "    + ", $_, "\n");
        }
    }

    close($rsync_h)
        or fault "Something went wrong during $pack filepack replication: $?";
    ## close temporary include file.
    close($tmpfh);

    if ($server->modified) {
        $server->display(1, "  - file replication completed.\n");
    }
    else {
        $server->display(1, "  - files already syncronized.\n");
    }

    ## TODO: run mod_cmd's if there was a modification

}

sub sync_apt_conf
{
    my $self   = shift->new;

    my $server = $self->server;

    return 1 if $self->get_key('no_apt_conf_sync');

    my ($pid, $rh, $wh) = $self->rw_command('sync-apt-conf');

    print $wh $server->apt_conf;
    close($wh);

    my $res = <$rh>;

    my $worked = 1 if (close($rh) and $res =~ /success/);

    waitpid $pid, 0;

    return $worked;
}

sub sync_apt_preferences
{
    my $self   = shift->new;

    my $server = $self->server;

    return 1 if $self->get_key('no_apt_preferences_sync');

    my ($pid, $rh, $wh) = $self->rw_command('sync-apt-preferences');

    print $wh $server->apt_preferences;
    close($wh);

    my $res = <$rh>;

    my $worked = 1 if (close($rh) and $res =~ /success/);

    waitpid $pid, 0;

    return $worked;
}

sub sync_apt_sources
{
    my $self   = shift->new;

    my $server = $self->server;

    return 1 if $self->get_key('no_apt_sources_sync');

    my ($pid, $rh, $wh) = $self->rw_command('sync-apt-sources');

    print $wh $server->apt_sources_list;
    close($wh);

    my $res = <$rh>;

    my $worked = 1 if (close($rh) and $res =~ /success/);

    waitpid $pid, 0;

    return $worked;
}

sub apt_get_update
{
    my $self = shift->new;
    return $self->gui_command('apt-get-update');
}

sub sync_debian_packages
{
    my $self   = shift->new;

    my $server = $self->server;

    $server->modified(0);

    my $v = $server->option('v');
    $server->display(1, "* debian package syncronization\n");

    $server->display(1, "  - determining server packages...");
    my $spkgs = $server->packages($v);
    $server->display(1, "done.\n");

    $server->display(1, "  - determining client packages...");
    my $cpkgs = $self->packages($v);
    $server->display(1, "done.\n");

    my $ipkgs = $self->ignore_debs;

    $server->display(1, "  - calculating differences...");
    my ($install, $remove, $purge) =
        $self->server->debian_package_diff($spkgs, $cpkgs, $ipkgs, $v);

    ## return if there is nothing to do.
    unless (@$install || @$remove || @$purge) {
        $server->display(1, "client up-to-date.\n");
        return;
    }

    $server->modified(1);

    ## have client update package sources.
    $server->display(1, "client out-of-sync.\n");
    $server->display(1, "  - syncronizing package sources...\n");

    $self->sync_apt_conf
        or fault("client sync-apt-conf returned status: $?");

    $self->sync_apt_preferences
        or fault("client sync-apt-preferences returned status: $?");

    $self->sync_apt_sources
        or fault("client sync-apt-sources returned status: $?");

    $self->apt_get_update
        or fault("client apt-get-update returned status: $?");

    $server->display(1, "  - installing/removing packages on client with apt-get...\n");

    my @apt = "package-install";
    push @apt, @$install; ## packages to install
    push @apt, map { "$_-" } @$remove;              ## packages to remove
    push @apt, map { "$_-" } @$purge;               ## packages to purge

    $self->gui_command(@apt)
            or fault("client apt-get-install returned status: $?");

    if (@$purge) {
        $server->display(1, "  - purging packages on client with apt-get...\n");
        my @apt = ("package-purge", @$purge);
        $self->gui_command(@apt)
                or fault("client apt-get-purge returned status: $?");
    }

    $server->display(1, "  - package syncronization completed!\n");
}

sub server
{
    my $self = shift->new;

    return $self->{server};
}

1;
